/* $NetBSD: $ */
/*-
  * Copyright (c) 2009 UPMC/LIP6
  * All rights reserved.
  * This software is distributed under the following condiions
  * compliant with the NetBSD foundation policy.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
  *
  * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  * POSSIBILITY OF SUCH DAMAGE.
  */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD: cpu.c,v 1.24 2005/12/11 12:18:39 christos Exp $");

#include "opt_multiprocessor.h"
#include "opt_cputype.h"        /* which mips CPUs do we support? */
#include "opt_ddb.h"
#include "opt_kgdb.h"

#define __INTR_PRIVATE
#include <sys/malloc.h>
#include <sys/types.h>
#include <machine/proc.h>
#include <sys/cpu.h>
#include <machine/intr.h>
#include <machine/xicu.h>
#include <machine/iopic.h>
#include <machine/boot_fdt.h>
#include <machine/db_machdep.h>
#include <ddb/db_output.h>
#include <sys/atomic.h>
#include <sys/kcpuset.h>

#include "vciiopic.h"

/*
 * map IPL to mask for cp0:status. bit 1 means interrupt is disabled
 * line 6 is unused and always masked.
 */

#define MIPS_SPL_NONE		(MIPS_INT_MASK_3 | MIPS_INT_MASK_4 | MIPS_INT_MASK_5)
#define MIPS_SPL_SOFTCLOCK	(MIPS_SPL_NONE | MIPS_SOFT_INT_MASK_0)
#define MIPS_SPL_SOFTBIO	(MIPS_SPL_SOFTCLOCK)
#define MIPS_SPL_SOFTNET	(MIPS_SPL_SOFTCLOCK | MIPS_SOFT_INT_MASK_1)
#define MIPS_SPL_SOFTSERIAL	(MIPS_SPL_SOFTNET)
#define MIPS_SPL_VM		(MIPS_SPL_SOFTSERIAL | MIPS_INT_MASK_2)
#define	MIPS_SPL_SCHED		(MIPS_SPL_VM | MIPS_INT_MASK_1)
#define MIPS_SPL_DDB		(MIPS_SPL_SCHED)
#define MIPS_SPL_HIGH		(MIPS_INT_MASK)

static const struct ipl_sr_map tsarmips_ipl_sr_map = {
    .sr_bits = {
	[IPL_NONE] = MIPS_SPL_NONE,
	[IPL_SOFTCLOCK] = MIPS_SPL_SOFTCLOCK,
	[IPL_SOFTBIO] = MIPS_SPL_SOFTBIO,
	[IPL_SOFTNET] = MIPS_SPL_SOFTNET,
	[IPL_SOFTSERIAL] = MIPS_SPL_SOFTSERIAL,
	[IPL_VM] = MIPS_SPL_VM,
	[IPL_SCHED] = MIPS_SPL_SCHED,
	[IPL_DDB] = MIPS_SPL_DDB,
	[IPL_HIGH] = MIPS_SPL_HIGH
    }
};

void
intr_init()
{
	ipl_sr_map = tsarmips_ipl_sr_map;
}

#ifdef MULTIPROCESSOR

void
cpu_broadcast_ipi(int tag)
{
	CPU_INFO_ITERATOR cii;
	struct cpu_info *ci;
	for (CPU_INFO_FOREACH(cii, ci)) {
		if ((ci->ci_flags & CPUF_RUNNING) != 0 && ci != curcpu()) {
			(void)cpu_send_ipi(ci, tag);
		}
	}
}

int
cpu_send_ipi(struct cpu_info *ci, int tag)
{
	return xicu_send_ipi(ci, tag);
}

/*
 * Halt this cpu
 */
void
cpu_halt(void)
{
	int index = cpu_index(curcpu());

	printf("cpu%d: shutting down\n", index);
	atomic_and_ulong(&curcpu()->ci_flags, ~CPUF_RUNNING);
	spl0();	 /* allow interrupts e.g. further ipi ? */
	for (;;) ;      /* spin */

	/* NOTREACHED */
}

/*
 * Pause this cpu
 */
void
cpu_pause(struct reg *regsp)
{
	int s = splhigh();
	for (;;) {
		atomic_or_ulong(&curcpu()->ci_flags, CPUF_PAUSED);
		do {
			;
		} while ((curcpu()->ci_flags & CPUF_PAUSED) != 0);

#if defined(DDB)
		if (ddb_running_on_this_cpu_p())
			cpu_Debugger();
		if (ddb_running_on_any_cpu_p())
			continue;
#endif
		break;
	}

	splx(s);
}

/*
 * Pause all running cpus, excluding current cpu.
 */
void
cpu_pause_others(void)
{
	extern int cold;
	int count = 0;
	u_long limit = curcpu()->ci_cpu_freq;   /* some finite amount of time */
	CPU_INFO_ITERATOR cii;
	struct cpu_info *ci;

	cpu_broadcast_ipi(IPI_SUSPEND);

	KASSERT(cold || limit > 0);
	while (limit--) {
		count = 0;
		for (CPU_INFO_FOREACH(cii, ci)) {
			if ((ci->ci_flags & CPUF_PAUSED) == 0 &&
			    ci != curcpu()) {
				count++;
			}
		}
		if (count == 0)
			break;
	}
	if (count != 0) {
		printf("Failed to pause: ");
		for (CPU_INFO_FOREACH(cii, ci)) {
			if ((ci->ci_flags & CPUF_PAUSED) == 0 &&
			    ci != curcpu()) {
				printf(" cpu%d", cpu_index(ci));
			}
		}
	}
}

/*
 * Resume a single cpu
 */
void
cpu_resume(int index)
{
	struct cpu_info *ci = cpu_lookup(index);
	KASSERT(ci != NULL);
	atomic_and_ulong(&ci->ci_flags, ~CPUF_PAUSED);

}

/*
 * Resume all paused cpus.
 */
void
cpu_resume_others(void)
{
	CPU_INFO_ITERATOR cii;
	struct cpu_info *ci;
	for (CPU_INFO_FOREACH(cii, ci)) {
		if (ci->ci_flags & CPUF_PAUSED) {
			atomic_and_ulong(&ci->ci_flags, ~CPUF_PAUSED);
		}
	}
}

int
cpu_is_paused(int index)
{
	struct cpu_info *ci;
	if (cpu_index(curcpu()) == index)
		ci = curcpu();
	else
		ci = cpu_lookup(index);
	KASSERT(ci != NULL);
	if (ci->ci_flags & CPUF_PAUSED)
		return 1;
	return 0;
}

#ifdef DDB
void
cpu_debug_dump(void)
{
	CPU_INFO_ITERATOR cii;
	struct cpu_info *ci;
	char present, running, paused;

	db_printf("CPU CPUID STATE CPUINFO	    CPL INT MTX IPIS\n");
	for (CPU_INFO_FOREACH(cii, ci)) {
		present = (ci->ci_flags & CPUF_PRESENT) ? 'H' : '-';
		running = (ci->ci_flags & CPUF_RUNNING) ? 'R' : '-';
		paused  = (ci->ci_flags & CPUF_PAUSED) ? 'P' : '-';
		db_printf("%3d 0x%03lx %c%c%c %p "
			"%3d %3d %3d "
			"0x%02lx/0x%02lx\n",
			cpu_index(ci), ci->ci_cpuid,
			running, present, paused,
			ci, ci->ci_cpl, ci->ci_idepth, ci->ci_mtx_count,
			ci->ci_active_ipis, ci->ci_request_ipis);
	}
}
#endif
#endif /* MULTIPROCESSOR */

void
cpu_intr(int pri, vaddr_t pc, u_int32_t status)
{
	xicu_intr(pri, pc, status);
}

irq_t
intr_establish(const struct boot_fdt_prop *irq, ipl_t ipl,
    const char *name, char *intrstr, int (*handler)(void *), void *ih_arg,
    struct cpu_info *ci)
{
	struct fdt_node *pic_node;
	device_t pic_dev;

	if (irq == NULL)
		return NULL;

	pic_node = fdt_find_node_from_ref(irq->prop_value[0]);
	if (pic_node == NULL)
		return NULL;

	pic_dev = pic_node->fdt_node_dev;
	if (pic_dev == NULL)
		return NULL;

	if (device_is_a(pic_dev, "xicu")) {
		return xicu_intr_establish(pic_dev, IRQ_HWI,
		    be32toh(irq->prop_value[1]), ipl, name, intrstr,
		    handler, ih_arg, ci);
	}

#if NVCIIOPIC > 0
	printf("intr_establish on %s\n", device_xname(pic_dev));
	if (device_is_a(pic_dev, "iopic")) {
		return iopic_intr_establish(pic_dev,
		    be32toh(irq->prop_value[1]), ipl, name, intrstr,
		    handler, ih_arg, ci);
	}
#endif
	return NULL;
}
